@@ -8,24 +8,28 @@ from copy import deepcopy
 from pathlib import Path
 from urllib.parse import urlsplit
 
-import ujson
-from aiohttp import ClientSession, TCPConnector
+import orjson
+from aiohttp import ClientSession, TCPConnector, AsyncResolver
+from tqdm import tqdm
 
 from .config.log import log_config
 from .config.operations import operations
+from .config.settings import trending_params
 from .constants import *
 from .login import login
-from .utils import find_key, build_query, get_headers
+from .utils import find_key, build_query, get_headers, set_qs
 
 try:
     if get_ipython().__class__.__name__ == 'ZMQInteractiveShell':
         import nest_asyncio
+
         nest_asyncio.apply()
 except:
     ...
 
 if platform.system() != 'Windows':
     import uvloop
+
     asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
 else:
     asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
@@ -37,7 +41,7 @@ logger = logging.getLogger(__name__)
 
 
 class Scraper:
-    GRAPHQL_URL = 'https://twitter.com/i/api/graphql/'
+    GRAPHQL_URL = 'https://api.twitter.com/graphql'
 
     def __init__(self, username: str, password: str):
         self.session = login(username, password)
@@ -116,7 +120,7 @@ class Scraper:
         return res
 
     async def process(self, urls: list, headers: dict) -> tuple:
-        conn = TCPConnector(limit=0, ssl=False, ttl_dns_cache=69)
+        conn = TCPConnector(limit=100, ssl=False, ttl_dns_cache=69, resolver=AsyncResolver())
         async with ClientSession(headers=headers, connector=conn) as s:
             # add cookies from logged-in session
             s.cookie_jar.update_cookies(self.session.cookies)
@@ -138,7 +142,7 @@ class Scraper:
             logger.debug(f'failed to download {url}: {e}')
 
     async def pagination(self, res: list, operation: tuple, limit: int) -> tuple:
-        conn = TCPConnector(limit=len(res), ssl=False, ttl_dns_cache=69)
+        conn = TCPConnector(limit=100, ssl=False, ttl_dns_cache=69, resolver=AsyncResolver())
         headers = get_headers(self.session)
         headers['content-type'] = "application/json"
         async with ClientSession(headers=headers, connector=conn) as s:
@@ -212,10 +216,15 @@ class Scraper:
             try:
                 r = await fn()
                 data = await r.json()
+                if r.status == 429:
+                    logger.debug(f'rate limit exceeded: {r.url}')
+                    return r, {}
+                if find_key(data, 'errors'):
+                    logger.debug(f'[{WARN}ERROR{RESET}]: {data}')
                 return r, data
             except Exception as e:
                 if i == retries:
-                    logger.debug(f'{WARN}Max retries exceeded{RESET}\n{e}')
+                    logger.debug(f'{WARN} Max retries exceeded{RESET}\n{e}')
                     return
                 t = 2 ** i + random.random()
                 logger.debug(f'retrying in {f"{t:.2f}"} seconds\t\t{e}')
@@ -226,8 +235,9 @@ class Scraper:
             for d in data:
                 path = Path(f'data/raw/{d[ID]}')
                 path.mkdir(parents=True, exist_ok=True)
-                with open(path / f'{time.time_ns()}_{name}.json', 'w') as fp:
-                    ujson.dump(d, fp, indent=4)
+                (path / f'{time.time_ns()}_{name}.json').write_text(
+                    orjson.dumps(d, option=orjson.OPT_INDENT_2).decode())
+
         except KeyError as e:
             logger.debug(f'failed to save data: {e}')
 
@@ -245,12 +255,16 @@ class Scraper:
         name = urlsplit(post_url).path.replace('/', '_')[1:]
         ext = urlsplit(cdn_url).path.split('/')[-1]
         try:
-            with open(f'{path}/{name}_{ext}', 'wb') as fp:
-                r = self.session.get(cdn_url, stream=True)
-                for chunk in r.iter_content(chunk_size=chunk_size):
-                    fp.write(chunk)
+            r = self.session.get(cdn_url, stream=True)
+            total_bytes = int(r.headers.get('Content-Length', 0))
+            desc = f'downloading: {name}'
+            with tqdm(total=total_bytes, desc=desc, unit='B', unit_scale=True, unit_divisor=1024) as pbar:
+                with open(f'{path}/{name}_{ext}', 'wb') as f:
+                    for chunk in r.iter_content(chunk_size=chunk_size):
+                        f.write(chunk)
+                        pbar.update(f.tell() - pbar.n)
         except Exception as e:
-            logger.debug(f'FAILED to download video: {post_url} {e}')
+            logger.debug(f'FAILED to download media: {post_url} {e}')
 
     def download_media(self, ids: list[int], photos: bool = True, videos: bool = True) -> None:
         res = self.tweet_by_rest_id(ids)
@@ -260,13 +274,35 @@ class Scraper:
             media = [y for x in find_key(r, 'media') for y in x]  # in case of arbitrary schema
             if photos:
                 photos = list({u for m in media if 'ext_tw_video_thumb' not in (u := m['media_url_https'])})
-                logger.debug(f'{photos = }')
+                # logger.debug(f'{photos = }')
                 if photos:
                     [self.download(url, photo) for photo in photos]
             if videos:
                 videos = [x['variants'] for m in media if (x := m.get('video_info'))]
                 hq_videos = {sorted(v, key=lambda d: d.get('bitrate', 0))[-1]['url'] for v in videos}
-                logger.debug(f'{videos = }')
-                logger.debug(f'{hq_videos = }')
+                # logger.debug(f'{videos = }')
+                # logger.debug(f'{hq_videos = }')
                 if hq_videos:
                     [self.download(url, video) for video in hq_videos]
+
+    def trends(self) -> dict:
+        """Get trends for all UTC offsets"""
+        url = set_qs('https://twitter.com/i/api/2/guide.json', trending_params)
+        headers = get_headers(self.session)
+        offsets = [f"{str(i).zfill(3)}00" if i < 0 else f"+{str(i).zfill(2)}00" for i in range(-12, 15)]
+        res = []
+        for offset in offsets:
+            headers['x-twitter-utcoffset'] = offset
+            r = self.session.get(url, headers=headers)
+            res.append(r.json())
+            logger.debug(f'getting trends for: {offset = }')
+        all_trends = {}
+        for data in res:
+            trends = find_key(data, 'item')
+            for t in trends:
+                all_trends |= {t['content']['trend']['name']: t}
+        path = Path(f'data/raw/trends')
+        path.mkdir(parents=True, exist_ok=True)
+        (path / f'{time.time_ns()}.json').write_text(
+            orjson.dumps(all_trends, option=orjson.OPT_INDENT_2).decode())
+        return all_trends
